#!/usr/bin/env python
# -*- encoding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function

from glob import glob
from os.path import basename, splitext

from setuptools import find_packages
from setuptools import setup

NAME = "electionguard"
VERSION = "1.0.0"
LICENSE = "MIT"
DESCRIPTION = "ElectionGuard: Support for e2e verified elections."
AUTHOR = "Microsoft Corporation"
AUTHOR_EMAIL = "electionguard@microsoft.com"
URL = "https://github.com/microsoft/electionguard-python"
PROJECT_URLS = {
    "Changelog": "https://github.com/microsoft/electionguard-python/blob/master/CHANGELOG.rst",
    "Issue Tracker": "https://github.com/microsoft/electionguard-python/issues",
}
CLASSIFIERS = [
    # http://pypi.python.org/pypi?%3Aaction=list_classifiers
    "Development Status :: 3 - Alpha",  # TODO Update when Stable
    "Intended Audience :: Developers",
    "License :: OSI Approved :: MIT License",
    "Operating System :: Unix",
    "Operating System :: POSIX",
    "Operating System :: MacOS",
    "Operating System :: Microsoft :: Windows",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: Implementation :: CPython",
    "Topic :: Utilities",
]

setup(
    name=NAME,
    version=VERSION,
    license=LICENSE,
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    url=URL,
    packages=find_packages("src"),
    package_dir={"": "src"},
    py_modules=[splitext(basename(path))[0] for path in glob("src/*.py")],
    include_package_data=True,
    zip_safe=False,
    classifiers=CLASSIFIERS,
    project_urls=PROJECT_URLS,
    python_requires=">=3.8",
    install_requires=["gmpy2>=2.0.8"],
)



from typing import List
from jsons import set_serializer
from os import mkdir, path
from jsons import set_serializer, dump


from .ballot_box import CiphertextBallot
from .decryption_mediator import PlaintextTally
from .election import CiphertextElectionContext, ElectionConstants, ElectionDescription
from .key_ceremony import CoefficientValidationSet
from .tally import CiphertextTally
from .group import ElementModP, ElementModQ

RESULTS_DIR = "results" + path.sep
COEFFICIENTS_DIR = RESULTS_DIR + "coefficients" + path.sep
BALLOTS_DIR = RESULTS_DIR + "encrypted_ballots" + path.sep
SPOILED_DIR = RESULTS_DIR + "spoiled_ballots" + path.sep

DESCRIPTION_FILE_NAME = "description"
CONTEXT_FILE_NAME = "context"
CONSTANTS_FILE_NAME = "constants"
ENCRYPTED_TALLY_FILE_NAME = "encrypted_tally"
TALLY_FILE_NAME = "tally"

COEFFICIENT_PREFIX = "coefficient_validation_set_"
BALLOT_PREFIX = "ballot_"


def publish(
    description: ElectionDescription,
    context: CiphertextElectionContext,
    constants: ElectionConstants,
    ciphertext_ballots: List[CiphertextBallot],
    ciphertext_tally: CiphertextTally,
    plaintext_tally: PlaintextTally,
    coefficient_validation_sets: List[CoefficientValidationSet] = None,
    results_directory: str = RESULTS_DIR,
) -> None:
    """Publishes the election record as json"""

    set_serializers()

    make_directory(results_directory)
    description.to_json_file(DESCRIPTION_FILE_NAME, results_directory)
    context.to_json_file(CONTEXT_FILE_NAME, results_directory)
    constants.to_json_file(CONSTANTS_FILE_NAME, results_directory)

    make_directory(COEFFICIENTS_DIR)
    if coefficient_validation_sets is not None:
        for coefficient_validation_set in coefficient_validation_sets:
            set_name = COEFFICIENT_PREFIX + coefficient_validation_set.owner_id
            coefficient_validation_set.to_json_file(set_name, COEFFICIENTS_DIR)

    make_directory(BALLOTS_DIR)
    for ballot in ciphertext_ballots:
        ballot_name = BALLOT_PREFIX + ballot.object_id
        ballot.to_json_file(ballot_name, BALLOTS_DIR)

    make_directory(SPOILED_DIR)
    for ballot in ciphertext_tally.spoiled_ballots.values():
        ballot_name = BALLOT_PREFIX + ballot.object_id
        ballot.to_json_file(ballot_name, SPOILED_DIR)

    ciphertext_tally.to_json_file(ENCRYPTED_TALLY_FILE_NAME, results_directory)
    plaintext_tally.to_json_file(TALLY_FILE_NAME, results_directory)


def make_directory(directory_path: str) -> None:
    """Create a directory only if it does not exist"""
    if not path.exists(directory_path):
        mkdir(directory_path)


def set_serializers() -> None:
    """Set serializers for jsons to use to cast specific classes"""
    set_serializer(lambda p, **_: str(p), ElementModP)
    set_serializer(lambda q, **_: str(q), ElementModQ)
    set_serializer(lambda tally, **_: dump(tally.cast), CiphertextTally)
    set_serializer(lambda tally, **_: dump(tally.contests), PlaintextTally)



from dataclasses import dataclass
from jsons import (
    KEY_TRANSFORMER_CAMELCASE,
    KEY_TRANSFORMER_SNAKECASE,
    dumps,
    loads,
    JsonsError,
)
from typing import cast, TypeVar, Generic

T = TypeVar("T")

JSON_FILE_EXTENSION: str = ".json"
WRITE: str = "w"
JSON_PARSE_ERROR = '{"error": "Object could not be parsed due to json issue"}'

# base10 as string for ElementModP and ElementModQ


@dataclass
class Serializable(Generic[T]):
    """
    Serializable class with methods to convert to json
    """

    def to_json(self) -> str:
        """
        Serialize to json
        :return: the json representation of this object
        """
        try:
            return cast(
                str,
                dumps(
                    self,
                    strip_privates=True,
                    strip_nulls=True,
                    key_transformer=KEY_TRANSFORMER_CAMELCASE,
                ),
            )
        except JsonsError:
            return JSON_PARSE_ERROR

    def to_json_file(self, file_name: str, file_path: str = "") -> None:
        """
        Serialize an object to a json file
        """
        write_json_file(self.to_json(), file_name, file_path)

    @classmethod
    def from_json(cls, data: str) -> T:
        """
        Deserialize the provided data string into the specified instance
        """
        return cast(T, loads(data, cls, key_transformer=KEY_TRANSFORMER_SNAKECASE))


def write_json_file(json_data: str, file_name: str, file_path: str = "") -> None:
    """
    Write json data string to json file
    """
    json_file_path: str = file_path + file_name + JSON_FILE_EXTENSION
    with open(json_file_path, WRITE) as json_file:
        json_file.write(json_data)
